
/******************************************************************************
   SyncWall: Wallpaper manager
   Copyright(C) 2011-2013  xbee@xbee.net

   This program is free software; you can redistribute it and/or
   modify it under the terms of the GNU General Public License
   as published by the Free Software Foundation; either version 2
   of the License, or (at your option) any later version.

   This program is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with this program; if not, write to the Free Software
   Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
 *******************************************************************************/

#include <QtCore/QDebug>
#include <QtGui/QPainter>

#include "ListImagesModel.h"
#include "ImageInfos.h"
#include "InfosUpdater.h"
#include "ThreadPool.h"

const int shadowBorderSize = 3;
const int whiteBorderSize = 3;
const int shadowStrength = 127;

bool lessThanByRow( const QModelIndex &indexA, const QModelIndex &indexB)
{
    return( indexA.row() > indexB.row() );
}

QPixmap ListImagesModel::generateFuzzyRect(const QSize& size, const QColor& color, int radius) const
{
    QPixmap pix(size);
    const QColor transparent(0, 0, 0, 0);
    pix.fill(transparent);

    QPainter painter(&pix);
    painter.setRenderHint(QPainter::Antialiasing, true);

    // Fill middle
    painter.fillRect(pix.rect().adjusted(radius, radius, -radius, -radius), color);

    // Corners
    QRadialGradient gradient;
    gradient.setColorAt(0, color);
    gradient.setColorAt(1, transparent);
    gradient.setRadius(radius);
    QPoint center;

    // Top Left
    center = QPoint(radius, radius);
    gradient.setCenter(center);
    gradient.setFocalPoint(center);
    painter.fillRect(0, 0, radius, radius, gradient);

    // Top right
    center = QPoint(size.width() - radius, radius);
    gradient.setCenter(center);
    gradient.setFocalPoint(center);
    painter.fillRect(center.x(), 0, radius, radius, gradient);

    // Bottom left
    center = QPoint(radius, size.height() - radius);
    gradient.setCenter(center);
    gradient.setFocalPoint(center);
    painter.fillRect(0, center.y(), radius, radius, gradient);

    // Bottom right
    center = QPoint(size.width() - radius, size.height() - radius);
    gradient.setCenter(center);
    gradient.setFocalPoint(center);
    painter.fillRect(center.x(), center.y(), radius, radius, gradient);

    // Borders
    QLinearGradient linearGradient;
    linearGradient.setColorAt(0, color);
    linearGradient.setColorAt(1, transparent);

    // Top
    linearGradient.setStart(0, radius);
    linearGradient.setFinalStop(0, 0);
    painter.fillRect(radius, 0, size.width() - 2 * radius, radius, linearGradient);

    // Bottom
    linearGradient.setStart(0, size.height() - radius);
    linearGradient.setFinalStop( 0, size.height() );
    painter.fillRect(radius, int( linearGradient.start().y() ), size.width() - 2 * radius, radius, linearGradient);

    // Left
    linearGradient.setStart(radius, 0);
    linearGradient.setFinalStop(0, 0);
    painter.fillRect(0, radius, radius, size.height() - 2 * radius, linearGradient);

    // Right
    linearGradient.setStart(size.width() - radius, 0);
    linearGradient.setFinalStop(size.width(), 0);
    painter.fillRect(int( linearGradient.start().x() ), radius, radius, size.height() - 2 * radius, linearGradient);
    return pix;
}

void ListImagesModel::drawShadow(QPainter* painter, const QRect& rect)  const
{
    const QPoint shadowOffset(-shadowBorderSize, -shadowBorderSize + 1);

    int key = rect.height() * 1000 + rect.width();

    QMap<int, QPixmap>::Iterator it = m_shadowCache.find(key);
    if ( it == m_shadowCache.end() )
    {
        QSize size = QSize(rect.width() + 2 * shadowBorderSize, rect.height() + 2 * shadowBorderSize);
        QColor color(0, 0, 0, shadowStrength);
        QPixmap shadow = generateFuzzyRect(size, color, shadowBorderSize);
        it = m_shadowCache.insert(key, shadow);
    }
    painter->drawPixmap( rect.topLeft() + shadowOffset, it.value() );
}

QPixmap ListImagesModel::shadoizePixmap(const QImage &pm) const
{
    QPixmap icon = QPixmap( pm.width() + 2 * shadowBorderSize + 2 * whiteBorderSize,pm.height() + 2 * shadowBorderSize + 2 * whiteBorderSize);

    QPainter painter(&icon);
    icon.fill( QColor(244,244,244) );

    QRect thumbnailRect = QRect(
        shadowBorderSize + whiteBorderSize,
        shadowBorderSize + whiteBorderSize - 1,
        pm.width() + whiteBorderSize,
        pm.height() + whiteBorderSize);

    drawShadow(&painter, thumbnailRect);

    // fill the "photo border"
    painter.fillRect(shadowBorderSize,shadowBorderSize,pm.width() + 2 * whiteBorderSize,pm.height() + 2 * whiteBorderSize,Qt::white);

    // fill the center pixmap
    painter.drawPixmap( shadowBorderSize + whiteBorderSize,shadowBorderSize + whiteBorderSize,QPixmap::fromImage(pm) );

    painter.end();

    return icon;
}

ListImagesModel::ListImagesModel(QObject *parent)
    : QAbstractItemModel( parent ),
    m_updater(NULL),
    m_mode(MODE_DETAILS),
    m_lastAppliedRow(-1),
    m_lastPreviewedRow(-1),
    m_nbScreens(1)
{
    // create updater infos
    m_updater = new InfosUpdater();

    // move the updater to this thread
    m_updater->moveToThread( ThreadPool::getInstance()->getThread() );

    qRegisterMetaType<QList<ImageInfos *> >("QList<ImageInfos *>");

    connect( m_updater,  SIGNAL( updated( ImageInfos * ) ), this, SLOT( refresh( ImageInfos *) ) );
    connect( m_updater,  SIGNAL( finished( ) ), this, SIGNAL( updateFinished() ) );

    connect( this, SIGNAL( updateInfos(const QList<ImageInfos *> &) ), m_updater, SLOT( check( const QList<ImageInfos *> & ) ) );
    connect( this, SIGNAL( storeInfos(ImageInfos *) ), m_updater, SLOT( storeInfos(ImageInfos *) ) );
    connect( this, SIGNAL( removeInfos(const QString &) ), m_updater, SLOT( removeInfos(const QString &) ) );
}

ListImagesModel::~ListImagesModel()
{
    m_updater->cancel();
    delete m_updater;
}

QVariant ListImagesModel::headerData (int inSection, Qt::Orientation inOrientation, int inRole) const
{
    if (inRole != Qt::DisplayRole || inOrientation != Qt::Horizontal || m_mode != MODE_DETAILS)
        return QVariant ();

    switch (inSection)
    {
        case COLUMN_NAME:
            return tr ("Name");

        case COLUMN_RATE:
            return tr ("Rate");

        case COLUMN_IMAGESIZE:
            return tr ("Image size");

        case COLUMN_FILESIZE:
            return tr ("File size");

        case COLUMN_USER_ADJUST:
            return tr ("Offset defined by user");
    }
    return QVariant();
}

QVariant ListImagesModel::data( const QModelIndex &index, int role ) const
{
    if( !index.isValid() )
        return QVariant();

    if (m_mode == MODE_DETAILS)
        return dataDetails(index, role);
    else
        return dataIcons(index, role);

    return QVariant();
}

QVariant ListImagesModel::dataDetails( const QModelIndex &index, int role ) const
{
    if( !index.isValid() )
        return QVariant();

    if( role == Qt::DisplayRole )
    {
        if (index.column() == COLUMN_NAME)
            return m_filesList.at( index.row() )->getFilename();
        else if (index.column() == COLUMN_RATE)
            return m_filesList.at( index.row() )->getRate();
        else if (index.column() == COLUMN_IMAGESIZE)
            return m_filesList.at( index.row() )->getImageSize();
        else if (index.column() == COLUMN_FILESIZE)
            return m_filesList.at( index.row() )->getFileSizeAsString();
        else if (index.column() == COLUMN_USER_ADJUST)
            return m_filesList.at( index.row() )->getOffsetAsString(m_nbScreens);
    }

    // "highlight" current wallpaper
    else if (role == Qt::BackgroundRole && index.row() == m_lastAppliedRow)
    {
        return QBrush ( QColor(255,200,0) );
    }

    return QVariant();
}

QVariant ListImagesModel::dataIcons( const QModelIndex &index, int role ) const
{
    if( !index.isValid() )
        return QVariant();

    if( role == Qt::DecorationRole )
        return shadoizePixmap( m_filesList.at( index.row() )->getThumbnail() );

    //return QPixmap::fromImage(m_filesList.at( index.row() )->getThumbnail());

    if ( role == Qt::ToolTipRole)
        return m_filesList.at( index.row() )->getFilename();

    if( role == Qt::DisplayRole )
        return m_filesList.at( index.row() )->getLabel();

    if( role == Qt::UserRole )
        return m_filesList.at( index.row() )->getImageSize();

    return QVariant();
}

bool ListImagesModel::setData(const QModelIndex &index, const QVariant &value, int role)
{
    if (!index.isValid() || role != Qt::EditRole)
        return false;

    if (index.column() == ListImagesModel::COLUMN_RATE)
    {
        m_filesList.at( index.row() )->setRate( value.toInt() );
        emit storeInfos( m_filesList.at( index.row() ) ); // save to db
        emit dataChanged(index, index); // update view
        // must recompute random indexes
        emit updateFinished();
        return true;
    }
    else if (index.column() == ListImagesModel::COLUMN_USER_ADJUST)
    {
        m_filesList.at( index.row() )->setOffset(value.toPoint(),m_nbScreens);
        emit storeInfos( m_filesList.at( index.row() ) ); // save to db
        emit dataChanged(index, index); // update view
        return true;
    }

    return false;
}

Qt::ItemFlags ListImagesModel::flags( const QModelIndex &index ) const
{
    if( !index.isValid() )
        return Qt::ItemIsDropEnabled;

    if (index.column() == ListImagesModel::COLUMN_RATE)
        return ( Qt::ItemIsEnabled | Qt::ItemIsSelectable | Qt::ItemIsEditable );
    else
        return ( Qt::ItemIsEnabled | Qt::ItemIsSelectable);
}

int ListImagesModel::rowCount( const QModelIndex &parent ) const
{
    if( parent.isValid() )
        return 0;

    return m_filesList.size();
}

void ListImagesModel::setDisplayMode(int mode)
{
    if (mode == m_mode)
        return;

    m_mode = mode;

    reset();
}

QModelIndex ListImagesModel::index(int row, int column, const QModelIndex &parent) const
{
    if ( !parent.isValid() ) // root item
    {
        if(row >= (int)m_filesList.size() || row < 0)
            return QModelIndex();

        return createIndex(row, column, m_filesList[row]);
    }

    return QModelIndex();
}

QModelIndex ListImagesModel::parent ( const QModelIndex & /*index*/ ) const
{
    return QModelIndex();
}

int ListImagesModel::columnCount ( const QModelIndex & /*parent*/  ) const
{
    switch (m_mode)
    {
        case MODE_DETAILS:
            return COLUMN_NB;
    }
    return 1;
}

void ListImagesModel::refresh (ImageInfos *ii)
{
    int index = m_filesList.indexOf(ii);
    if (index >= 0)
    {
        QModelIndex mIndex = createIndex(index, 0, m_filesList[index]);
        emit dataChanged( mIndex,mIndex.sibling(0,0) );
    }
}

ImageInfos *ListImagesModel::getInfos(const QString &filename)
{
    // TO BE ENHANCED
    foreach (ImageInfos * ii, m_filesList)
    {
        if (ii->getFilename() == filename)
            return ii;
    }
    return NULL;
}

void ListImagesModel::addImages( const QStringList & list )
{
    int row = m_filesList.size();

    // check for double
    QStringList valid;
    QList<ImageInfos *> oldIIList; // store already defined file (just update infos)
    foreach (const QString &file, list)
    {
        if ( !m_alreadyDefined.contains(file) )
        {
            valid << file;
            m_alreadyDefined << file;
        }
        else
        {
            oldIIList << getInfos(file);
        }
    }

    beginInsertRows( QModelIndex(), row, row + valid.size() );

    QList<ImageInfos *> newIIList;
    foreach( const QString &location, valid )
    newIIList << new ImageInfos(location);
    m_filesList.append(newIIList);

    newIIList << oldIIList; // update also existing file (just in case)

    emit updateInfos(newIIList);

    endInsertRows();
}

void ListImagesModel::removeRows(const QModelIndexList& indexes)
{
    // we must sort the index by order
    QModelIndexList sorted(indexes);
    qSort(sorted.begin(),sorted.end(),lessThanByRow);

    foreach(const QModelIndex &index, sorted)
    {
        if (index.row() == m_lastAppliedRow)
            m_lastAppliedRow = -1;
        removeRows(index.row(),1);
    }
}

bool ListImagesModel::removeRows( int row, int count, const QModelIndex &parent )
{
    if( parent.isValid() )
        return false;

    if( row >= m_filesList.size() || row + count <= 0 )
        return false;

    int beginRow = qMax( 0, row );
    int endRow = qMin( row + count - 1, m_filesList.size() - 1 );

    beginRemoveRows( parent, beginRow, endRow );

    while( beginRow <= endRow )
    {
        ImageInfos *ii = m_filesList.takeAt( beginRow );
        emit removeInfos( ii->getFilename() );
        m_alreadyDefined.remove( ii->getFilename() ); // update map
        delete ii;
        ++beginRow;
    }

    endRemoveRows();

    // recompute random indexes
    emit updateFinished();
    return true;
}

void ListImagesModel::clear()
{
    m_updater->cancel();
    while( m_updater->isRunning() )
        ;
    qDeleteAll(m_filesList);
    m_filesList.clear();
    reset();
}

int ListImagesModel::getIndex(const QModelIndex &index) const
{
    if ( index.isValid() )
        return index.row();

    return -1;
}

int ListImagesModel::getIndex(const QString &filename) const
{
    // TO BE ENHANCED
    for(int i = 0; i < m_filesList.size(); i++)
        if (m_filesList[i]->getFilename() == filename)
            return i;


    return -1;
}

int ListImagesModel::getRating(int row) const
{
    if ( row >= 0 && row < m_filesList.size() )
        return m_filesList[row]->getRate();

    return 0;
}

QString ListImagesModel::getFilename(int row) const
{
    if ( row >= 0 && row < m_filesList.size() )
        return m_filesList[row]->getFilename();

    return QString();
}

QString ListImagesModel::getFilename(const QModelIndex &index) const
{
    if ( index.isValid() )
        return m_filesList[index.row()]->getFilename();

    return QString();
}

QPoint ListImagesModel::getOffset(int row) const
{
    if ( row >= 0 && row < m_filesList.size() )
        return m_filesList[row]->getOffset(m_nbScreens);

    return QPoint();
}

QPoint ListImagesModel::getOffset(const QModelIndex &index) const
{
    if ( index.isValid() )
        return m_filesList[index.row()]->getOffset(m_nbScreens);

    return QPoint();
}

void ListImagesModel::setLastAppliedRow(int ind)
{
    if (ind == m_lastAppliedRow)
        return;

    // update display
    QModelIndex prevIndex = createIndex(m_lastAppliedRow, 0);
    m_lastAppliedRow = ind;
    QModelIndex newIndex = createIndex(m_lastAppliedRow, 0);

    if ( prevIndex.isValid() )
        emit dataChanged( prevIndex,prevIndex.sibling(0,0) );
    if ( newIndex.isValid() )
        emit dataChanged( newIndex,newIndex.sibling(0,0) );
}

void ListImagesModel::setLastPreviewedRow(int ind)
{
    m_lastPreviewedRow = ind;
}

void ListImagesModel::resetChache()
{
    m_updater->cancel();
    m_updater->clearCache();
    foreach(ImageInfos * ii, m_filesList)
    ii->reset();
    emit updateInfos(m_filesList);
}

void ListImagesModel::setScreenNumber(int num)
{
    if (num == m_nbScreens)
        return;

    m_nbScreens = num;
    reset(); // must take all new config for new screen number
}
