
/******************************************************************************
   SyncWall: Wallpaper manager
   Copyright(C) 2011-2013  xbee@xbee.net

   This program is free software; you can redistribute it and/or
   modify it under the terms of the GNU General Public License
   as published by the Free Software Foundation; either version 2
   of the License, or (at your option) any later version.

   This program is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with this program; if not, write to the Free Software
   Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
 *******************************************************************************/

#include <QtCore/QTimer>
#include <QtNetwork/QHostInfo>
#include <QtNetwork/QUdpSocket>

#include "NetworkManager.h"
#include "LogHandler.h"
#include "AppSettings.h"

const int defaultUdpPort = 55555;
const int defaultPingTimeout = 15 * 1000;

NetworkManager::NetworkManager(QObject *parent)
    : QObject(parent),
    m_udpSocket(NULL),
    m_timer(NULL),
    m_pingTimer(NULL),
    m_udpPort(0),
    m_mode(0),
    m_desc("no description"),
    m_waitAnswer(NULL),
    m_initialized(false)
{
    loadSettings();

    // can't initialize other member because current thread isn't final thread !!!
}

NetworkManager::~NetworkManager()
{
    //saveSettings();
    //delete m_udpSocket;

    m_timer->stop();
    m_waitAnswer->stop();
}

void NetworkManager::initialize()
{
    if (m_initialized)
        return;

    m_timer = new QTimer(this);
    m_pingTimer = new QTimer(this);
    m_waitAnswer = new QTimer(this);

    m_udpSocket = new QUdpSocket();

    connect( m_udpSocket, SIGNAL( readyRead() ),this, SLOT( processPendingDatagrams() ) );
    connect( m_waitAnswer, SIGNAL( timeout() ), this, SIGNAL( serverDown() ), Qt::QueuedConnection );
    connect( m_pingTimer, SIGNAL( timeout() ), this, SLOT( pingServer() ) );

    m_initialized = true;
}

void NetworkManager::start()
{
    initialize();

    updateConfig();
    if (m_mode != MODE_SERVER)
        pingServer();
}

void NetworkManager::loadSettings()
{
    AppSettings settings;

    settings.beginGroup("NetworkManager");

    m_udpPort = settings.value("udpPort",defaultUdpPort).toInt();
    m_serverName = settings.value("serverName","").toString();
    m_mode = settings.value("mode",MODE_SERVER).toInt();

    settings.endGroup();
}

void NetworkManager::saveSettings()
{
    AppSettings settings;

    settings.beginGroup("NetworkManager");

    settings.setValue("udpPort", m_udpPort);
    settings.setValue("serverName", m_serverName);
    settings.setValue("mode", m_mode);

    settings.endGroup();
}

void NetworkManager::setServerName(const QString &name)
{
    m_serverName = name;
}

void NetworkManager::updateConfig()
{
    int port;

    // listen to different port switch network mode
    if (m_mode == MODE_SERVER)
    {
        port = m_udpPort;
        m_pingTimer->stop();
        LogHandler::getInstance()->reportDebug( QObject::tr("NetworkManager: Mode SERVER listen to %1 port").arg(port) );
    }
    else
    {
        port = m_udpPort + 1;

        // start the "ping" server timer
        m_pingTimer->start(defaultPingTimeout);
        LogHandler::getInstance()->reportDebug( QObject::tr("NetworkManager: Mode CLIENT listen to %1 port").arg(port) );
    }

    m_udpSocket->disconnectFromHost();
    m_udpSocket->bind(port);
}

void NetworkManager::setPort(int p)
{
    if (p == m_udpPort)
        return;

    m_udpPort = p;

    updateConfig();
}

void NetworkManager::setMode(int mode)
{
    if (m_mode == mode)
        return;

    m_mode = mode;

    // clear client list
    if (m_mode == MODE_CLIENT)
        m_clients.clear();

    updateConfig();
}

void NetworkManager::broadcastDatagramSendServerIdentification(const QHostAddress &addr)
{
    int port;

    if (m_mode == MODE_SERVER)
        port = m_udpPort + 1;
    else
        port = m_udpPort;

    LogHandler::getInstance()->reportDebug( QObject::tr("NetworkManager: Send Server Identification for \"%1\" on port %2").arg( addr.toString() ).arg(port) );

    // prepare identification and description
    QByteArray datagram = QString("Syncwall Server Answer desc=#%2#").arg(m_desc).toLatin1();
    m_udpSocket->writeDatagram(datagram.data(), datagram.size(),addr, port);
}

void NetworkManager::processPendingDatagrams()
{
    while ( m_udpSocket->hasPendingDatagrams() )
    {
        QByteArray datagram;
        datagram.resize( m_udpSocket->pendingDatagramSize() );
        QHostAddress sender;
        quint16 senderPort;
        m_udpSocket->readDatagram(datagram.data(), datagram.size(),&sender, &senderPort);

        LogHandler::getInstance()->reportDebug( QObject::tr("NetworkManager: Received message on port %1").arg(senderPort) );

        if ( datagram.startsWith("Syncwall Server Call") )
        {
            LogHandler::getInstance()->reportDebug( QObject::tr("NetworkManager: Received Server call datagram: \"%1\"").arg( datagram.data() ) );
            broadcastDatagramSendServerIdentification(sender);

            if ( !m_clients.contains(sender) )
            {
                LogHandler::getInstance()->reportDebug( QObject::tr("NetworkManager: Adding %1 to client list").arg( sender.toString() ) );
                m_clients << sender;
            }
            else
            {
                LogHandler::getInstance()->reportDebug( QObject::tr("NetworkManager: %1 is already in the client list").arg( sender.toString() ) );
            }

            // send the current background
            broadcastDatagramSendBackground(sender);
        }
        else if ( datagram.startsWith("Syncwall Server Answer") )
        {
            m_waitAnswer->stop();
            emit serverAlive();
        }
        else if ( datagram.startsWith("Syncwall Change Background") )
        {
            QList<QByteArray> bg = datagram.split(' ');
            LogHandler::getInstance()->reportDebug( QObject::tr("NetworkManager: Received background change: \"%1\"").arg( QString(bg[3]) ) );
            emit newMessage( QString(bg[3]) );
        }
    }
}

void NetworkManager::searchServer(const QString &hostname)
{
    QHostInfo info = QHostInfo::fromName(hostname);
    if ( !info.addresses().isEmpty() )
    {
        broadcastDatagramSearchServers(info.addresses()[0]);
        m_waitAnswer->start(2000);
    }
    else
        LogHandler::getInstance()->reportError( QObject::tr("NetworkManager: no address found for host %1 (%2)").arg(hostname).arg( info.errorString() ) );
}

void NetworkManager::broadcastDatagramSearchServers(const QHostAddress &addr)
{
    LogHandler::getInstance()->reportDebug( QObject::tr("NetworkManager: Send Server Call for \"%1\" on port %2").arg( addr.toString() ).arg(m_udpPort) );
    QByteArray datagram = "Syncwall Server Call";
    m_udpSocket->writeDatagram(datagram.data(), datagram.size(), addr, m_udpPort);
}

void NetworkManager::broadcastDatagramSendBackground(const QHostAddress &addr)
{
    LogHandler::getInstance()->reportDebug( QObject::tr("NetworkManager: Send message to \"%1\" (%2) on port %3").arg( addr.toString() ).arg(m_currentMessage).arg(m_udpPort + 1) );
    QByteArray datagram = QString("Syncwall Change Background %1").arg(m_currentMessage).toLatin1();
    m_udpSocket->writeDatagram(datagram.data(), datagram.size(), addr, m_udpPort + 1);
}

void NetworkManager::sendNewMessage(const QString &img)
{
    m_currentMessage = img;

    for (int i = 0; i < m_clients.size(); i++)
    {
        LogHandler::getInstance()->reportDebug( QObject::tr("NetworkManager:: Sending new message to client %1 ").arg( m_clients[i].toString() ) );
        broadcastDatagramSendBackground(m_clients[i]);
    }
}

QAbstractSocket::SocketState NetworkManager::getState() const
{
    return m_udpSocket->state();
}

void NetworkManager::pingServer()
{
    LogHandler::getInstance()->reportDebug( QObject::tr("NetworkManager:: Ping server %1").arg(m_serverName) );
    searchServer(m_serverName);
}
