
/******************************************************************************
   SyncWall: Wallpaper manager
   Copyright(C) 2011-2013  xbee@xbee.net

   This program is free software; you can redistribute it and/or
   modify it under the terms of the GNU General Public License
   as published by the Free Software Foundation; either version 2
   of the License, or (at your option) any later version.

   This program is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with this program; if not, write to the Free Software
   Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
 *******************************************************************************/

#include <QtCore/QTimer>
#include <QtNetwork/QTcpSocket>
#include <QtCore/QFile>

#include "TcpClient.h"
#include "LogHandler.h"

static quint32 const PAYLOADSIZE = 65536;

//status strings
static const QString INACTIVE_STRING = "inactive";
static const QString LISTENING_STRING = "listening";
static const QString CONNECTING_STRING = "connecting";
static const QString NEGOTIATING_STRING = "negotiating";
static const QString SENDING_STRING = "sending";
static const QString RECEIVING_STRING = "receiving";
static const QString FINISHED_STRING = "finished";
static const QString UNKNOWN_STRING = "unknown";

//---------------------------------------------------------------------------------------------
TcpClient::TcpClient(QObject *parent) :
    QObject(parent),
    m_file(0)
{
    m_socket = new QTcpSocket(this);

    m_timeOutCounter = 0;
    m_connectionTimer = new QTimer(this);

    QObject::connect( m_socket, SIGNAL( connected() ),
                      this, SLOT( on_socketConnected() ) );

    QObject::connect( m_socket, SIGNAL( bytesWritten(qint64) ),
                      this, SLOT( updateClientProgress(qint64) ) );

    QObject::connect( m_socket, SIGNAL( disconnected() ),
                      this, SLOT( on_socketDisconnected() ) );

    QObject::connect( m_connectionTimer, SIGNAL( timeout() ),
                      this, SLOT( timeout() ) );
}

//---------------------------------------------------------------------------------------------
TcpClient::~TcpClient()
{
    delete m_socket;
}

//---------------------------------------------------------------------------------------------
QString TcpClient::getStateAsString(quint8 state)
{
    QString state_string = "";

    switch(state)
    {
        case INACTIVE:
            state_string = INACTIVE_STRING;
            break;

        case CONNECTING:
            state_string = CONNECTING_STRING;
            break;

        case NEGOTIATING:
            state_string = NEGOTIATING_STRING;
            break;

        case SENDING:
            state_string = SENDING_STRING;
            break;

        case FINISHED:
            state_string = FINISHED_STRING;
            break;

        default:
            state_string = UNKNOWN_STRING;
    }

    return state_string;
}

//---------------------------------------------------------------------------------------------
void TcpClient::setState(quint8 new_state)
{
    m_currentState = new_state;
    LogHandler::getInstance()->reportInfo( tr("Client: State changed to '%1'").arg( getStateAsString(new_state) ) );

    switch(m_currentState)
    {
        case INACTIVE:

            m_bytesWritten = 0;
            m_blockSize = 0;
            m_connectionTimer->stop();
            m_tempBytesWritten = 0;
            m_timeOutCounter = 0;

            if(m_socket)
                m_socket->abort();

            if( m_file && m_file->isOpen() )
                m_file->close();

            break;

        case CONNECTING:

            break;

        case NEGOTIATING:
            break;

        case SENDING:
            break;
    }

    emit stateChanged(new_state);
}

//---------------------------------------------------------------------------------------------
void TcpClient::timeout()
{
    quint32 speed = (m_bytesWritten - m_tempBytesWritten) >> 10;
    m_tempBytesWritten = m_bytesWritten;

    if(speed > 0)
    {
        m_timeOutCounter = 0;
    }
    else
    {
        if(++m_timeOutCounter == m_timeOut)
        {
            LogHandler::getInstance()->reportInfo( tr("Client: Connection timed out") );
            setState(INACTIVE);
        }
    }
}

//---------------------------------------------------------------------------------------------
void TcpClient::updateClientProgress(qint64 num_bytes)
{
    if(m_currentState != SENDING)
        return;

    m_bytesWritten += (int)num_bytes;

    if(m_bytesToWrite > 0)
    {
        m_fileBlock = m_file->read(PAYLOADSIZE);
        m_bytesToWrite -= (int)m_socket->write( m_fileBlock.left( qMin(m_bytesToWrite, PAYLOADSIZE) ) );
    }

    if(m_bytesWritten >= m_fileSize)
    {
        setState(FINISHED);
        LogHandler::getInstance()->reportInfo( tr("Client: File transfer complete") );
        setState(INACTIVE);
    }
}

//---------------------------------------------------------------------------------------------
void TcpClient::on_socketConnected()
{
    LogHandler::getInstance()->reportInfo( tr("Client: Connection to server established") );
    setState(NEGOTIATING);

    sendFileInformation();

    QObject::connect( m_socket, SIGNAL( readyRead() ),
                      this, SLOT( readMessage() ) );
}

//---------------------------------------------------------------------------------------------
void TcpClient::sendFileInformation()
{
    QString filepath; // = filename_textbox_->text();
    QString filename = filepath;
    QFile file(filepath);

    if( !file.exists() )
    {
        LogHandler::getInstance()->reportError( tr("Client: File '%1' does not exist").arg(filepath) );
        setState(INACTIVE);
        return;
    }

    //extract filename only (without the path)
    int index = filename.lastIndexOf('/');

    if(index != -1)
        filename = filename.right(filename.length() - index - 1);

    //send the file information
    QByteArray block;
    QDataStream out(&block, QIODevice::WriteOnly);
    quint32 filesize = file.size();

    out.setVersion(QDataStream::Qt_4_0);
    out << (quint16)0;
    out << (quint8)'I' << filename << filesize;
    out.device()->seek(0);
    out << (quint16)( block.size() - sizeof(quint16) );

    m_socket->write(block);

    m_connectionTimer->start(1000);

    LogHandler::getInstance()->reportInfo( tr("Client: File information sent to server") );
}

//---------------------------------------------------------------------------------------------
void TcpClient::sendFile()
{
    QString filepath; // = filename_textbox_->text();
    m_file = new QFile(filepath);

    if( !m_file->open(QIODevice::ReadOnly) )
    {
        LogHandler::getInstance()->reportError( tr("Client: Error reading file '%1'").arg(filepath) );
        setState(INACTIVE);
        return;
    }

    m_timeOutCounter = 0;
    m_tempBytesWritten = 0;

    //m_connectionTimer->start(1000);

    LogHandler::getInstance()->reportInfo( tr("Client: Starting file transfer ...") );
    setState(SENDING);

    m_fileSize = m_file->size();
    m_fileBlock = m_file->read(PAYLOADSIZE);

    m_bytesToWrite = m_fileSize - (int)m_socket->write(m_fileBlock);
}

//---------------------------------------------------------------------------------------------
void TcpClient::readMessage()
{
    QDataStream in(m_socket);
    in.setVersion(QDataStream::Qt_4_0);

    if(m_blockSize == 0)
    {
        if( m_socket->bytesAvailable() < (int)sizeof(quint16) )
            return;

        in >> m_blockSize;
    }

    if(m_socket->bytesAvailable() < m_blockSize)
        return;

    quint8 message_type;
    in >> message_type;

    if(m_currentState == NEGOTIATING)
    {
        switch(message_type)
        {
            case 'A': //Peer accepted the file
                LogHandler::getInstance()->reportInfo( tr("Client: Sever accepted the file") );
                sendFile();
                break;

            default:
                if(m_currentState != NEGOTIATING)
                    LogHandler::getInstance()->reportError( tr("Received unknown message type: %1").arg(message_type) );
        }
    }
}

//---------------------------------------------------------------------------------------------
void TcpClient::on_socketError(int socket_error)
{
    LogHandler::getInstance()->reportError( tr("Unable to connect: %1.").arg( m_socket->errorString() ) );

    setState(INACTIVE);
}

//---------------------------------------------------------------------------------------------
void TcpClient::on_socketDisconnected()
{
    if( (m_currentState == NEGOTIATING || m_currentState == SENDING) )
    {
        LogHandler::getInstance()->reportError( tr("Client: Connection closed. Transfer aborted.") );
    }

    setState(INACTIVE);
}
