
/******************************************************************************
   SyncWall: Wallpaper manager
   Copyright(C) 2011-2012  xbee@xbee.net

   This program is free software; you can redistribute it and/or
   modify it under the terms of the GNU General Public License
   as published by the Free Software Foundation; either version 2
   of the License, or (at your option) any later version.

   This program is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with this program; if not, write to the Free Software
   Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
 *******************************************************************************/

#include "TcpServer.h"
#include "LogHandler.h"

#include <QtNetwork/QHostAddress>
#include <QtNetwork/QTcpSocket>

static quint32 const PAYLOADSIZE = 65536;

//status strings
static const QString INACTIVE_STRING = "inactive";
static const QString LISTENING_STRING = "listening";
static const QString CONNECTING_STRING = "connecting";
static const QString NEGOTIATING_STRING = "negotiating";
static const QString SENDING_STRING = "sending";
static const QString RECEIVING_STRING = "receiving";
static const QString FINISHED_STRING = "finished";
static const QString UNKNOWN_STRING = "unknown";

//---------------------------------------------------------------------------------------------
TcpServer::TcpServer(QObject *parent)
    : QObject(parent)
{
    m_server = new QTcpServer(parent);
    m_connectionTimer = new QTimer(this);
    m_file = 0;
    m_socket = 0;
    m_port = 0;
    m_timeOutCounter = 0;
    m_timeOut = 10;  //FIXXXME: Should be set by the user

    QObject::connect( m_server, SIGNAL( newConnection() ),
                      this, SLOT( on_server__newConnection() ) );

    QObject::connect( m_connectionTimer, SIGNAL( timeout() ),
                      this, SLOT( timeout() ) );

    m_currentState = 0;
    m_reverseConnection = false;
}

//---------------------------------------------------------------------------------------------
TcpServer::~TcpServer()
{
    if(m_server)
        delete m_server;
}

//---------------------------------------------------------------------------------------------
QString TcpServer::getStateAsString(quint8 state)
{
    QString state_string = "";

    switch(state)
    {
        case INACTIVE:
            state_string = INACTIVE_STRING;
            break;

        case LISTENING:
            state_string = LISTENING_STRING;
            break;

        case CONNECTING:
            state_string = CONNECTING_STRING;
            break;

        case NEGOTIATING:
            state_string = NEGOTIATING_STRING;
            break;

        case RECEIVING:
            state_string = RECEIVING_STRING;
            break;

        case FINISHED:
            state_string = FINISHED_STRING;
            break;

        default:
            state_string = UNKNOWN_STRING;
    }

    return state_string;
}

//---------------------------------------------------------------------------------------------
bool TcpServer::start(quint16 port)
{
    m_port = port;

    if(m_reverseConnection)
    {
        LogHandler::getInstance()->reportInfo( tr("TcpServer: Initiating reverse connection") );

        if(!m_socket)
            m_socket = new QTcpSocket(this);

        QObject::connect( m_socket, SIGNAL( connected() ),
                          this, SLOT( on_socketConnected() ) );
        QObject::connect( m_socket, SIGNAL( readyRead() ),
                          this, SLOT( readMessage() ) );
        QObject::connect( m_socket, SIGNAL( disconnected() ),
                          this, SLOT( on_socketDisconnected() ) );

        QString target; // = parent_->getTargetReceiveCombobox()->lineEdit()->text();

        //if(parent_->isIP(target))
        //   m_socket->connectToHost(QHostAddress(target), m_port);
        //else
        m_socket->connectToHost(target, m_port);
    }
    else
    {
        while( !m_server->isListening() && !m_server->listen(QHostAddress::Any, m_port) )
        {
            LogHandler::getInstance()->reportInfo( tr("TcpServer: Unable to start the server: %1").arg( m_server->errorString() ) );
            return false;
        }

        LogHandler::getInstance()->reportInfo( tr("TcpServer: Starting server...") );
    }

    m_blockSize = 0;
    m_bytesReceived = 0;
    m_tempBytesReceived = 0;

    setState(LISTENING);
    return true;
}

//---------------------------------------------------------------------------------------------
void TcpServer::stop()
{
    if( m_server->isListening() )
        m_server->close();

    setState(INACTIVE);
    LogHandler::getInstance()->reportInfo( tr("TcpServer: TcpServer stopped") );
}

//---------------------------------------------------------------------------------------------
void TcpServer::timeout()
{
    quint32 speed = (m_bytesReceived - m_tempBytesReceived) >> 10;
    m_tempBytesReceived = m_bytesReceived;

    if(speed > 0)
    {
        m_timeOutCounter = 0;
    }
    else
    {
        if(++m_timeOutCounter == m_timeOut)
        {
            LogHandler::getInstance()->reportInfo( tr("TcpServer: Connection timed out") );
            abort();
        }
    }
}

//---------------------------------------------------------------------------------------------
void TcpServer::abort()
{
    if( m_currentState == RECEIVING && m_file->exists() )
        m_file->remove();

    stop();
}

//---------------------------------------------------------------------------------------------
QString TcpServer::errorString() const
{
    return m_server->errorString();
}

//---------------------------------------------------------------------------------------------
void TcpServer::setState(quint8 new_state)
{
    m_currentState = new_state;
    LogHandler::getInstance()->reportInfo( tr("TcpServer: State changed to '%1'").arg( getStateAsString(new_state) ) );

    switch(m_currentState)
    {
        case INACTIVE:
            if(m_socket)
                m_socket->abort();

            if(m_file)
            {
                m_file->close();
                delete m_file;
                m_file = 0;
            }

            m_saveFilename = "";
            m_fileSize = 0;
            m_blockSize = 0;

            m_connectionTimer->stop();
            m_tempBytesReceived = 0;
            m_timeOutCounter = 0;

            break;

        case RECEIVING:

            break;

        default:
            break;
    }

    emit stateChanged(new_state);
}

//---------------------------------------------------------------------------------------------
void TcpServer::on_server__newConnection()
{
    if(m_socket)
        delete m_socket;

    m_socket = m_server->nextPendingConnection();
    LogHandler::getInstance()->reportInfo( tr("TcpServer: %1 is connecting...").arg( m_socket->peerAddress().toString() ) );
    setState(CONNECTING);

    QObject::connect( m_socket, SIGNAL( readyRead() ),
                      this, SLOT( readMessage() ) );

    QObject::connect( m_socket, SIGNAL( disconnected() ),
                      this, SLOT( on_socketDisconnected() ) );
}

//---------------------------------------------------------------------------------------------
void TcpServer::on_socketDisconnected()
{
    if(m_currentState == RECEIVING)
    {
        LogHandler::getInstance()->reportInfo( tr("TcpServer: Connection closed. Transfer aborted.") );
        abort();
    }
}

//---------------------------------------------------------------------------------------------
void TcpServer::updateTcpServerProgress()
{
    m_bytesReceived += (int)m_socket->bytesAvailable();
    m_fileData.clear();
    m_fileData.resize(PAYLOADSIZE);
    m_fileData = m_socket->readAll();

    m_file->write(m_fileData);

    if(m_bytesReceived == m_fileSize)
    {
        setState(FINISHED);
        m_socket->close();
        m_file->flush();
        m_file->close();
        delete m_file;
        m_file = 0;

        LogHandler::getInstance()->reportInfo( tr("TcpServer: File transfer complete") );
        stop();
    }
}

//---------------------------------------------------------------------------------------------
void TcpServer::readMessage()
{
    QDataStream in(m_socket);
    in.setVersion(QDataStream::Qt_4_0);

    if(m_blockSize == 0)
    {
        if( m_socket->bytesAvailable() < (int)sizeof(quint16) )
            return;

        in >> m_blockSize;
    }

    if(m_socket->bytesAvailable() < m_blockSize)
        return;

    if(m_currentState == CONNECTING)
    {
        setState(NEGOTIATING);

        quint8 messageType;
        QString filename;
        QString filesizeString;

        in >> messageType;

        switch(messageType)
        {
            case 'I': //File info
            {
                in >> filename >> m_fileSize;

                if(m_fileSize < 1024)
                    filesizeString = tr("%1 Byte").arg(m_fileSize);
                else if(m_fileSize < 1048576)
                    filesizeString = tr("%1 KB").arg(static_cast<double>(m_fileSize) / 1024, 0, 'f', 2);
                else if(m_fileSize < 1073741824)
                    filesizeString = tr("%1 MB").arg(static_cast<double>(m_fileSize) / 1048576, 0, 'f', 2);
                else
                    filesizeString = tr("%1 GB").arg(static_cast<double>(m_fileSize) / 1073741824, 0, 'f', 2);

                LogHandler::getInstance()->reportInfo( tr("TcpServer: Received file information:") );
                LogHandler::getInstance()->reportInfo( tr("        '%1', size: %2").arg(filename).arg(filesizeString) );

//             m_saveFilename = QFileDialog::getSaveFileName(parent_, "Save file as...",
//                              last_dir_.append(filename), "");
                m_file = new QFile(m_saveFilename);

                if( !m_file->open(QIODevice::WriteOnly) )
                {
                    LogHandler::getInstance()->reportError( tr("Error writing to file. Transfert aborted.") );
                    setState(INACTIVE);
                }

                sendAcceptMessage();

                m_fileData.resize(PAYLOADSIZE);
                QObject::connect( m_socket, SIGNAL( readyRead() ),this, SLOT( updateTcpServerProgress() ) );

                m_timeOutCounter = 0;
                m_connectionTimer->start(1000);
                setState(RECEIVING);

                break;
            }

            default:

                LogHandler::getInstance()->reportError( tr("Received unknown message type: %1").arg(messageType) );
                break;
        }
    }
}

//---------------------------------------------------------------------------------------------
void TcpServer::sendAcceptMessage()
{
    QByteArray block;
    QDataStream out(&block, QIODevice::WriteOnly);

    out.setVersion(QDataStream::Qt_4_0);
    out << (quint16)0;
    out << (quint8)'A';
    out.device()->seek(0);
    out << (quint16)( block.size() - sizeof(quint16) );

    m_socket->write(block);
    LogHandler::getInstance()->reportInfo( tr("TcpServer: Accept message sent") );
}

//---------------------------------------------------------------------------------------------
void TcpServer::setTimeOut(quint16 time_out)
{
    m_timeOut = time_out;
}

//---------------------------------------------------------------------------------------------
void TcpServer::setReverseConnection(bool reverse_connection)
{
    m_reverseConnection = reverse_connection;
}

//---------------------------------------------------------------------------------------------
void TcpServer::on_socketConnected()
{
    LogHandler::getInstance()->reportInfo( tr("TcpServer: Reverse connection to %1 established").arg( m_socket->peerAddress().toString() ) );
    setState(CONNECTING);
}

//---------------------------------------------------------------------------------------------
bool TcpServer::getReverseConnection()
{
    return m_reverseConnection;
}
