#include <QtCore/QTimer>
#include <QtNetwork/QTcpSocket>

#include "tcpclient.h"
#include "core/log/loghandler.h"

using namespace Spy;


const int connectionTimeOut = 10000;

#define MSG_SIZE_ALIGN(size) (((size) + 1023) & ~1023)


static bool check_header(int at, int c)
{
	static const char prefix[] = "DATA";
	if (at < MSG_LEN_OFFSET)
		return (c == prefix[at]);
	return isdigit(c);
}

static bool check_header2(int at1, int at2)
{
	static const char prefix[] = "DATA";
	if ((at1 < MSG_LEN_OFFSET) && (at2 < MSG_LEN_OFFSET))
		return (prefix[at1] == prefix[at2]);
	if ((at1 < MSG_LEN_OFFSET) && (at2 >= MSG_LEN_OFFSET))
		return isdigit(prefix[at1]);
	if ((at1 >= MSG_LEN_OFFSET) && (at2 < MSG_LEN_OFFSET))
		return isdigit(prefix[at2]);
	return true;
}

TcpClient::TcpClient(QObject *parent) :
    QObject(parent),
    m_host("localhost"),
    m_port(45454)
{
    m_socket = new QTcpSocket(this);
	m_buffer.capacity = MSG_SIZE_ALIGN(MSG_HEADER_SIZE);
	m_buffer.base = new char[m_buffer.capacity];
	memset(m_buffer.base,0,m_buffer.capacity);
	m_buffer.start = m_buffer.base; 
	m_buffer.state = READ_HEADER;
	m_buffer.accept = 0;
	m_buffer.payload = 0;

	initMessageHeader();

    m_connectionTimer = new QTimer(this);
    m_connectionTimer->setInterval(connectionTimeOut);
    m_connectionTimer->setSingleShot(true);

    QObject::connect( m_socket, SIGNAL( connected() ),
                      this, SLOT( on_socketConnected() ) );

    QObject::connect( m_socket, SIGNAL( disconnected() ),
                      this, SLOT( on_socketDisconnected() ) );

    QObject::connect( m_connectionTimer, SIGNAL( timeout() ),
                      this, SLOT( timeout() ) );
}

TcpClient::~TcpClient()
{
	if (m_socket != NULL)
		delete m_socket;
	delete[] m_buffer.base;
}

void TcpClient::start(const QString &host, int port)
{
    qDebug() << "TcpClient::start " << host << " " << port;
    m_host = host;
    m_port = port;
    restart();
}

void TcpClient::restart()
{
    qDebug() << "TcpClient::restart";
    stop();
    m_connectionTimer->start();
    m_socket->connectToHost(m_host, m_port);
}

void TcpClient::stop()
{
    m_connectionTimer->stop();
    m_socket->abort();
}


void TcpClient::timeout()
{
    LogHandler::getInstance()->reportInfo( tr("Client: Connection to server timed out => restarting") );
    restart();
}

void TcpClient::on_socketConnected()
{
	qDebug() << "####client connected";
    LogHandler::getInstance()->reportInfo( tr("Client: Connection to server established") );
	emit stateChanged(true);  //--CB
    //setState(NEGOTIATING);

    QObject::connect( m_socket, SIGNAL( readyRead() ),
                      this, SLOT( readMessage() ) );
    m_connectionTimer->stop();
    //sendAcceptMessage();
}

void TcpClient::readMessage()
{
	bool match;
	int bytes = 0;
	while (m_socket->bytesAvailable() > 0)
	{
		if (m_buffer.state == READ_HEADER)
		{
			if (m_buffer.accept < MSG_HEADER_SIZE) 
			{
				if ((bytes = m_socket->read(m_buffer.start,MSG_HEADER_SIZE)) < 0)
					break;
				while ((bytes > 0) && m_buffer.accept < MSG_HEADER_SIZE)
					{
					if ((match = check_header(m_buffer.accept,*m_buffer.start)))
						m_header.val[m_buffer.accept] = *m_buffer.start;
					while (!match && (m_buffer.accept >= 0))
					{
						m_buffer.accept = m_header.kmp[m_buffer.accept];
						match = check_header(m_buffer.accept,*m_buffer.start);
					}
					m_buffer.accept++;
					m_buffer.start++;
					bytes--;
				}
			}
			if (m_buffer.accept == MSG_HEADER_SIZE)
			{
				m_buffer.payload = QString::fromUtf8(m_header.val+MSG_LEN_OFFSET,MSG_HEADER_SIZE-MSG_LEN_OFFSET).toInt();
				if (m_buffer.capacity < m_buffer.payload)
				{
					char* old = m_buffer.base;
					m_buffer.capacity = MSG_SIZE_ALIGN(m_buffer.payload);
					m_buffer.base = new char[m_buffer.capacity];
					memcpy(m_buffer.base,old,m_buffer.start-old);
					delete[] old;
				}
				if (m_buffer.payload > 0)
					m_buffer.state = READ_PAYLOAD;
				if (bytes > 0)
					memmove(m_buffer.base,m_buffer.start,bytes);
				m_buffer.start = m_buffer.base + bytes; 
				m_buffer.accept = qMin(m_buffer.payload,bytes);
			}
		}
		if (m_buffer.state == READ_PAYLOAD)
		{
			if (m_buffer.accept < m_buffer.payload)
			{
				int remaining = m_buffer.payload - m_buffer.accept;
				if ((bytes = m_socket->read(m_buffer.start,remaining)) < 0)
					break;
				m_buffer.start += bytes;
				m_buffer.accept += bytes;
			}
			if (m_buffer.accept == m_buffer.payload)
			{
				QString message = QString::fromUtf8(m_buffer.base,m_buffer.payload);
				//qDebug() << "#######message:" << message;
				emit messageReceived(message);
				m_buffer.state = READ_HEADER;
				int bytes = m_buffer.start - (m_buffer.base + m_buffer.payload);
				if (bytes > 0)
					memmove(m_buffer.base,m_buffer.start,bytes);
				m_buffer.start = m_buffer.base + bytes; 
				m_buffer.accept = 0;
			}
		}
	}	
}

void TcpClient::sendRawMessage(const RawMessage &mess)
{
    QByteArray block;
    QDataStream out(&block, QIODevice::WriteOnly);

    //out.setVersion(QDataStream::Qt_4_0);
    out << mess;

    m_socket->write(block);
    m_socket->flush();

    //qDebug() << "Client send " << block.size();
    qint64 offset = out.device()->pos(); // current offset in s
    //qDebug() << "Client datastream offset " << offset;

	emit messageEmitted(mess.data);
}

void TcpClient::send(const Message& message)
{
	try 
	{
		RawMessage raw_msg;
		raw_msg.data = message.toXml();
		sendRawMessage(raw_msg);
	} 
	catch (const ErrorType& err) 
	{
		Spy::LogHandler::getInstance()->reportError(QString("!!! Error found on <%1> : %2").arg(err.logged.origin).arg(err.logged.message)); 
	}
}

void TcpClient::on_socketError(int )
{
    LogHandler::getInstance()->reportError( tr("Unable to connect: %1.").arg( m_socket->errorString() ) );
	emit stateChanged(false);  //--CB
    //setState(INACTIVE);
}

void TcpClient::on_socketDisconnected()
{
	LogHandler::getInstance()->reportError( tr("Client: Connection closed"));
	emit stateChanged(false);  //--CB	
	/*
    if( (m_currentState == NEGOTIATING ) )
    {
        LogHandler::getInstance()->reportError( tr("Client: Connection closed. Transfer aborted.") );
    }
    setState(INACTIVE);
	*/
}

void TcpClient::initMessageHeader()
{
	int check = -1;
	m_header.kmp[0] = -1;
	for (int iter=0; iter < MSG_HEADER_SIZE-1; iter++)
	{
		while (check >=0 && !check_header2(iter,check))
			check = m_header.kmp[check];
		check++;
		if (check_header2(iter+1,check))
			m_header.kmp[iter+1] = m_header.kmp[check];
		else
			m_header.kmp[iter+1] = check;
	}
	memset(m_header.val,0,sizeof(m_header.val));
}