
#include "tcpserver.h"
#include "core/log/loghandler.h"

#include <QtNetwork/QHostAddress>
#include <QtNetwork/QTcpSocket>

using namespace Spy;

//status strings
static const QString INACTIVE_STRING = "inactive";
static const QString LISTENING_STRING = "listening";
static const QString CONNECTING_STRING = "connecting";
static const QString NEGOTIATING_STRING = "negotiating";
static const QString CONNECTED_STRING = "connected";
static const QString FINISHED_STRING = "finished";
static const QString UNKNOWN_STRING = "unknown";

const  int scheduleTime = 50; 

//---------------------------------------------------------------------------------------------
TcpServer::TcpServer(QObject *parent)
    : QObject(parent)
{
    m_server = new QTcpServer(parent);
    m_connectionTimer = new QTimer(this);
	m_scheduler = new QTimer(this);
	m_scheduler->setInterval(scheduleTime);
	connect(m_scheduler,SIGNAL(timeout()), this, SLOT(sendScheduledMessage()));
    m_socket = 0;
    m_port = 0;
    m_timeOutCounter = 0;
    m_timeOut = 10;
	
	memset(m_buffer,0,10000);

    QObject::connect( m_server, SIGNAL( newConnection() ),
                      this, SLOT( serverNewConnection() ) );

    QObject::connect( m_connectionTimer, SIGNAL( timeout() ),
                      this, SLOT( timeout() ) );

    m_currentState = 0;
}

//---------------------------------------------------------------------------------------------
TcpServer::~TcpServer()
{
    if(m_server)
        delete m_server;
}

//---------------------------------------------------------------------------------------------
QString TcpServer::getStateAsString(quint8 state)
{
    QString state_string = "";

    switch(state)
    {
    case INACTIVE:
        state_string = INACTIVE_STRING;
        break;

    case CONNECTED:
        state_string = CONNECTED_STRING;
        break;

    default:
        state_string = UNKNOWN_STRING;
    }

    return state_string;
}

//---------------------------------------------------------------------------------------------
bool TcpServer::start(quint16 port)
{
    m_port = port;

    while( !m_server->isListening() && !m_server->listen(QHostAddress::Any, m_port) )
    {
        LogHandler::getInstance()->reportInfo( tr("TcpServer: Unable to start the server: %1").arg( m_server->errorString() ) );
        return false;
    }

    LogHandler::getInstance()->reportInfo( tr("TcpServer: Starting server...") );

    m_bytesReceived = 0;
    m_tempBytesReceived = 0;

    setState(LISTENING);
    return true;
}

//---------------------------------------------------------------------------------------------
void TcpServer::stop()
{
    if( m_server->isListening() )
        m_server->close();

    setState(INACTIVE);
    LogHandler::getInstance()->reportInfo( tr("TcpServer: TcpServer stopped") );
}

//---------------------------------------------------------------------------------------------
void TcpServer::timeout()
{
    quint32 speed = (m_bytesReceived - m_tempBytesReceived) >> 10;
    m_tempBytesReceived = m_bytesReceived;

    if(speed > 0)
    {
        m_timeOutCounter = 0;
    }
    else
    {
        if(++m_timeOutCounter == m_timeOut)
        {
            LogHandler::getInstance()->reportInfo( tr("TcpServer: Connection timed out") );
            abort();
        }
    }
}

//---------------------------------------------------------------------------------------------
void TcpServer::abort()
{
    stop();
}

//---------------------------------------------------------------------------------------------
QString TcpServer::errorString() const
{
    return m_server->errorString();
}

//---------------------------------------------------------------------------------------------
void TcpServer::setState(quint8 new_state)
{
	if (m_currentState == new_state)
		return;
    m_currentState = new_state;
    LogHandler::getInstance()->reportInfo( tr("TcpServer: State changed to '%1'").arg( getStateAsString(new_state) ) );

    switch(m_currentState)
    {
    case INACTIVE:
        if(m_socket)
            m_socket->abort();

        m_blockSize = 0;

        m_connectionTimer->stop();
        m_tempBytesReceived = 0;
        m_timeOutCounter = 0;

        break;

    case LISTENING:
        break;

    case CONNECTED:
        break;
    }

    emit stateChanged(new_state);
}

//---------------------------------------------------------------------------------------------
void TcpServer::serverNewConnection()
{
    if(m_socket)
        delete m_socket;

    m_socket = m_server->nextPendingConnection();
    LogHandler::getInstance()->reportInfo( tr("TcpServer: %1 is connecting...").arg( m_socket->peerAddress().toString() ) );
    //--CBsetState(CONNECTING);
	setState(CONNECTED);

    QObject::connect( m_socket, SIGNAL( readyRead() ), this, SLOT( onReceipt() ) );
    QObject::connect( m_socket, SIGNAL( disconnected() ), this, SLOT( on_socketDisconnected() ) );
}

//---------------------------------------------------------------------------------------------
void TcpServer::on_socketDisconnected()
{
    if(m_currentState == CONNECTED)
    {
        LogHandler::getInstance()->reportInfo( tr("TcpServer: Disconnection Client side. Transfer aborted.") );

        // we consider there will be only one client for the moment
        // when this one disconnect, the server must listen any possible client reconnection
        setState(LISTENING);
    }
}

//---------------------------------------------------------------------------------------------

void TcpServer::onReceipt()
{
    if (m_receivedStarted == false) 
	{
        /* There it's a new message we are receiving.
           To start receiving it we must know its length, i.e. the 2 first bytes */
        if (m_socket->bytesAvailable() < 8) // DATA id + lenght
            return;

        /* Ok now we can start */
        m_receivedStarted = true;

        /* getting the length of the message */
		char data[5];
		memset(data,0,5);
		m_socket->read(data,4);
		m_socket->read(data,4);
		QString sizeStr(data);
		m_remainingLength = sizeStr.toInt();
        /* Recursive call to write less code =) */
        onReceipt();
    } 
	else 
	{
        /* Checking if the command is complete! */
        if (m_socket->bytesAvailable() >= m_remainingLength) 
		{
			m_connectionTimer->stop();

			m_socket->read(m_buffer,m_remainingLength);
			RawMessage mess;
			mess.size = m_remainingLength;
			mess.data = m_buffer;

			emit messageReceived(mess.data);

            m_receivedStarted = false;

            /* Recursive call to spare code =), there may be still data pending */
            onReceipt();
        }
    }
}

//---------------------------------------------------------------------------------------------

void TcpServer::readMessage()
{
    //setState(CONNECTED);

    //qDebug() << "server receive " << m_socket->bytesAvailable ();

    QDataStream in(m_socket);

    //in.setVersion(QDataStream::Qt_4_0);

    RawMessage mess;
    in >> mess;
    emit messageReceived(mess.data);
}

void TcpServer::sendRawMessage(const RawMessage &mess)
{
	if (!m_socket)
		return;


	if (!m_scheduler->isActive())
		m_scheduler->start();

	m_messagesToSend << mess;
}

void TcpServer::sendScheduledMessage()
{
	if (m_messagesToSend.isEmpty())
	{
		m_scheduler->stop();
		return;
	}

	if (!m_socket)
		return;


	RawMessage mess = m_messagesToSend.takeFirst();

	QByteArray block;
	QDataStream out(&block, QIODevice::WriteOnly);

	out.setVersion(QDataStream::Qt_4_0);
	out << mess;

	m_socket->write(block);
	m_socket->flush();

	//qint64 offset = out.device()->pos(); // current offset in s

	emit messageEmitted(mess.data);
}

//---------------------------------------------------------------------------------------------
void TcpServer::setTimeOut(quint16 time_out)
{
    m_timeOut = time_out;
}
